# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Services for users and user roles/groups."""

from .groups import GroupsService, GroupsServiceConfig
from .users import UsersService, UsersServiceConfig

__all__ = (
    "GroupsService",
    "GroupsServiceConfig",
    "UsersService",
    "UsersServiceConfig",
)



# -*- coding: utf-8 -*-
#
# Copyright (C)      2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Proxy models for user preferences."""


class NotificationPreferencesProxy:
    """Proxy for a user that allows mapping the form to the user object."""

    _notification_preferences = lambda self: self._user.preferences.get(
        "notifications", {}
    )

    def __init__(self, user):
        """."""
        super().__setattr__("_user", user)

    def __getattr__(self, attr):
        """."""
        return self._notification_preferences.get(attr, None)

    def __setattr__(self, attr, value):
        """."""
        self._user.preferences = {
            **self._user.preferences,
            "notifications": {
                **(self._notification_preferences()),
                attr: value,
            },
        }

    def __hasattr__(self, attr):
        """."""
        return attr in self._notification_preferences



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-Vocabularies is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users search facets definitions."""

from invenio_i18n import gettext as _
from invenio_records_resources.services.records.facets import TermsFacet

email_domain = TermsFacet(
    field="email.domain",
    label=_("Email domain"),
)

affiliations = TermsFacet(
    field="profile.affiliations.keyword",
    label=_("Affiliations"),
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
# Copyright (C) 2022-2023 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module providing management APIs for users and roles/groups."""

__version__ = "3.0.2"

__all__ = ("__version__",)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-users-resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Tests for user moderation.

These tests are separate from the service because this module will test collateral behaviours from moderating users. Examples:

- Locking mechanisms (involves user service methods and celery tasks)
"""

import time
from unittest.mock import MagicMock

import pytest
from invenio_access.permissions import system_identity
from invenio_cache.lock import CachedMutex

from invenio_users_resources.proxies import current_actions_registry
from invenio_users_resources.services.users.lock import ModerationMutex


def test_moderation_callbacks_success(
    user_service, user_res, user_moderator, monkeypatch, clear_cache
):
    """Test moderation actions (post block / restore)."""

    mocked_method = MagicMock(return_value=True)
    monkeypatch.setitem(current_actions_registry, "block", [mocked_method])
    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked

    # Verify that the callback was called once
    mocked_method.assert_called_once()


def test_moderation_callbacks_failure(
    user_service, user_res, user_moderator, monkeypatch, clear_cache
):
    """Test moderation actions (post block).

    This aims to test multiple behaviours:
    - If a callback fails, no changes are committed.
    - Even if a callback fails, the user block is committed.
    """

    def _block_action_success(user_id, uow=None):
        """Action to execute after blocking a user."""
        user_service.approve(system_identity, user_id, uow=uow)

    def _block_action_failure(user_id, uow=None):
        """Action raises an exception."""
        raise Exception("Callback failed")

    monkeypatch.setitem(
        current_actions_registry,
        "block",
        [_block_action_success, _block_action_failure],
    )
    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked

    # Registered callbacks will fail and their result won't be committed.
    ur = user_service.read(user_moderator.identity, user_res.id)
    assert ur.data["active"] == False
    assert "blocked_at" in ur.data


def test_moderation_callbacks_lock(
    app, user_service, user_res, user_moderator, monkeypatch, clear_cache
):
    """Tests the 'simplest' flow for user moderation in terms of locks (e.g. mutex).

    This test validates the basic flow of user moderation in terms of locks. The scenario involves the following steps:

        1. A user is blocked by a moderator.
        2. An action is executed after blocking the user (nothing happens).
        3. The test checks that the user's lock is removed by the celery task, indicating successful moderation.
        4. The test verifies that the user's data is consistent (user is approved and not blocked).
    """

    mocked_method = MagicMock(return_value=True)
    monkeypatch.setitem(
        current_actions_registry,
        "block",
        [mocked_method],
    )

    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked

    # Verify the callback was called. After this, the lock should be released.
    mocked_method.assert_called_once()

    # If the lock does not exist, it was removed by the celery task
    lock = ModerationMutex(user_res.id)
    assert not lock.exists()


@pytest.mark.parametrize(
    "default_timeout, renewal_timeout, expected_lock_state",
    [(1, 10, True), (2, 2, False)],
)
def test_moderation_callbacks_lock_renewal(
    app,
    user_service,
    user_res,
    user_moderator,
    monkeypatch,
    default_timeout,
    renewal_timeout,
    expected_lock_state,
    clear_cache,
):
    """Tests whether the lock is renewed after moderating a user.

    The test scenario involves the following steps:
        1. A user is moderated, and a lock is created for a specific `default_timeout` period.
        2. A celery task runs, renewing the lock timeout for a `renewal_timeout` period.
        3. The action inside the celery task simulates a process that takes time (sleeps for `default_timeout` seconds).
        4. The test validates whether the lock state matches the expected state after the renewal.


    An example with numbers:

        1) lock is created for 1 second.
        2) lock is renewed for 10 seconds.
        3) task sleeps for 1 second.
        4) if the lock exists, then it was renewed for 10 seconds.
    """

    def _block_action_success(user_id, uow=None):
        """Action to execute after blocking a user."""
        time.sleep(default_timeout)

        lock = ModerationMutex(user_id)
        assert lock.exists() == expected_lock_state

    monkeypatch.setitem(
        current_actions_registry,
        "block",
        [_block_action_success],
    )

    monkeypatch.setitem(
        app.config,
        "USERS_RESOURCES_MODERATION_LOCK_DEFAULT_TIMEOUT",
        default_timeout,
    )
    monkeypatch.setitem(
        app.config,
        "USERS_RESOURCES_MODERATION_LOCK_RENEWAL_TIMEOUT",
        renewal_timeout,
    )
    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Permission generators for users and groups."""


from invenio_records.dictutils import dict_lookup
from invenio_records_permissions.generators import (
    ConditionalGenerator,
    Generator,
    UserNeed,
)
from invenio_search.engine import dsl


class IfPublic(ConditionalGenerator):
    """Generator for different permissions based on the visibility settings."""

    def __init__(self, field_name, then_, else_, **kwargs):
        """Constructor."""
        super().__init__(then_=then_, else_=else_, **kwargs)
        self._field_name = field_name

    def _condition(self, record=None, **kwargs):
        """Condition to choose generators set."""
        if record is None:
            return False

        visibility = record.preferences.get(self._field_name, "restricted")
        is_public = visibility == "public"

        return is_public

    def query_filter(self, **kwargs):
        """Filters for queries."""
        field = f"preferences.{self._field_name}"
        q_public = dsl.Q("match", **{field: "public"})
        q_restricted = dsl.Q("match", **{field: "restricted"})
        then_query = self._make_query(self.then_, **kwargs)
        else_query = self._make_query(self.else_, **kwargs)

        if then_query and else_query:
            return (q_public & then_query) | (q_restricted & else_query)
        elif then_query:
            return q_public & then_query
        elif else_query:
            return q_public | (q_restricted & else_query)
        else:
            return q_public


class IfPublicEmail(IfPublic):
    """Generator checking the 'email_visibility' setting."""

    def __init__(self, then_, else_):
        """Constructor."""
        super().__init__("email_visibility", then_, else_)


class IfPublicUser(IfPublic):
    """Generator checking the 'visibility' setting."""

    def __init__(self, then_, else_):
        """Constructor."""
        super().__init__("visibility", then_, else_)


class Self(Generator):
    """Requires the users themselves."""

    def needs(self, record=None, **kwargs):
        """Set of Needs granting permission."""
        if record is not None:
            return [UserNeed(record.id)]

        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for the current identity."""
        if identity is not None:
            for need in identity.provides:
                if need.method == "id":
                    return dsl.Q("term", id=need.value)

        return []


class IfGroupNotManaged(ConditionalGenerator):
    """Generator for managed group access."""

    def __init__(self, then_, else_, **kwargs):
        """Constructor."""
        super().__init__(then_=then_, else_=else_, **kwargs)
        self._field_name = "is_managed"

    def _condition(self, record, **kwargs):
        """Condition to choose generators set."""
        if record is None:
            return False

        is_managed = dict_lookup(record, self._field_name)
        return not is_managed

    def query_filter(self, **kwargs):
        """Filters for queries."""
        q_all = dsl.Q("match_all")
        q_not_managed = dsl.Q("match", **{self._field_name: False})
        then_query = self._make_query(self.then_, **kwargs)
        else_query = self._make_query(self.else_, **kwargs)

        identity = kwargs.get("identity", None)

        if identity:
            for need in self.needs(**kwargs):
                if need in identity.provides:
                    return q_all & else_query

        return q_not_managed & then_query



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User service tests."""

import pytest
from invenio_access.permissions import system_identity
from invenio_records_resources.resources.errors import PermissionDeniedError


def test_groups_sort(app, groups, group_service):
    # default sort by name
    res = group_service.search(system_identity).to_dict()
    assert res["sortBy"] == "name"
    assert res["hits"]["total"] > 0
    hits = res["hits"]["hits"]
    assert hits[0]["id"] == "hr-dep"
    assert hits[1]["id"] == "it-dep"


def test_groups_no_facets(app, group, group_service):
    """Make sure certain fields ARE searchable."""
    res = group_service.search(system_identity)
    # if facets were enabled but not configured the value would be {}
    assert res.aggregations is None


def test_groups_fixed_pagination(app, groups, group_service):
    res = group_service.search(system_identity, params={"size": 1, "page": 2})
    assert res.pagination.page == 1
    assert res.pagination.size == 10


@pytest.mark.parametrize(
    "query",
    [
        # cannot search on title because is never set
        # see TODO in parse_role_data
        "id:it-dep",
        "name:IT Department",
        "+name:it",
        "IT",
    ],
)
def test_groups_search_field(app, group, group_service, query):
    """Make sure certain fields ARE searchable."""
    res = group_service.search(system_identity, q=query)
    assert res.total > 0


def test_groups_search(app, groups, group_service, user_pub, anon_identity):
    """Test group search."""

    # System can retrieve all groups.
    res = group_service.search(system_identity).to_dict()
    assert res["hits"]["total"] == len(groups)

    # Authenticated user can retrieve unmanaged groups
    res = group_service.search(user_pub.identity).to_dict()
    assert res["hits"]["total"] == len([g for g in groups if not g.is_managed])

    # Anon does not have permission to search
    with pytest.raises(PermissionDeniedError):
        group_service.search(anon_identity).to_dict()


def test_groups_read(app, groups, group_service, user_pub, anon_identity):
    """Test group read."""

    from invenio_accounts.models import Role

    for g in groups:
        # System can retrieve all groups.
        group_service.read(system_identity, g.name).to_dict()

        # Authenticated user can retrieve unmanaged groups
        if g.is_managed:
            with pytest.raises(PermissionDeniedError):
                group_service.read(user_pub.identity, g.name).to_dict()
        else:
            group_service.read(user_pub.identity, g.name).to_dict()

        # Anon does not have permission to search
        with pytest.raises(PermissionDeniedError):
            group_service.read(anon_identity, g.name).to_dict()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User and user group schemas."""

from invenio_access.permissions import system_user_id
from invenio_accounts.profiles.schemas import (
    validate_locale,
    validate_timezone,
    validate_visibility,
)
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.records.schema import (
    BaseGhostSchema,
    BaseRecordSchema,
)
from marshmallow import Schema, ValidationError, fields
from marshmallow_utils.fields import ISODateString, SanitizedUnicode
from marshmallow_utils.permissions import FieldPermissionsMixin


class UserPreferencesSchema(Schema):
    """Schema for user preferences."""

    visibility = fields.String(validate=validate_visibility)
    email_visibility = fields.String(validate=validate_visibility)
    locale = fields.String(validate=validate_locale)
    timezone = fields.String(validate=validate_timezone)


class UserProfileSchema(Schema):
    """Schema for user profiles."""

    full_name = fields.String()
    affiliations = fields.String()


class UserSchema(BaseRecordSchema, FieldPermissionsMixin):
    """Schema for users."""

    field_dump_permissions = {
        "email": "read_email",
        "created": "read_details",
        "updated": "read_details",
        "revision_id": "read_details",
        "active": "read_details",
        "confirmed": "read_details",
        "preferences": "read_details",
        "blocked_at": "read_system_details",
        "verified_at": "read_system_details",
        "confirmed_at": "read_system_details",
    }

    # NOTE: API should only deliver users that are active & confirmed
    active = fields.Boolean()
    confirmed = fields.Boolean(dump_only=True)
    is_current_user = fields.Method("is_self", dump_only=True)

    email = fields.String()
    username = fields.String()
    profile = fields.Dict()
    preferences = fields.Nested(UserPreferencesSchema)

    blocked_at = ISODateString()
    verified_at = ISODateString()
    confirmed_at = ISODateString()

    def is_self(self, obj):
        """Determine if identity is the current identity."""
        current_identity = self.context["identity"]

        _id = obj.get("id") or obj.id

        return (
            _id is not None
            and current_identity.id is not None
            and str(_id) == str(current_identity.id)
        )


class GroupSchema(BaseRecordSchema):
    """Schema for user groups."""

    name = fields.String()
    title = fields.String()
    description = fields.String()
    provider = fields.String(dump_only=True)
    is_managed = fields.Boolean(dump_only=True)


class UserGhostSchema(BaseGhostSchema):
    """User ghost schema."""

    id = SanitizedUnicode(dump_only=True)
    profile = fields.Constant(
        {
            "full_name": _("Deleted user"),
        },
        dump_only=True,
    )
    username = fields.Constant(_("Deleted user"), dump_only=True)


class SystemUserSchema(BaseGhostSchema):
    """System user schema."""

    id = fields.Constant(system_user_id, dump_only=True)
    profile = fields.Constant(
        {
            "full_name": _("System"),
        },
        dump_only=True,
    )
    username = fields.Constant(_("System"), dump_only=True)


class NotificationPreferences(Schema):
    """Schema for notification preferences."""

    enabled = fields.Bool()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User specific filters for notifications."""

from invenio_notifications.services.filters import RecipientFilter


class UserPreferencesRecipientFilter(RecipientFilter):
    """Recipient filter for notifications being enabled at all."""

    def __call__(self, notification, recipients):
        """Filter recipients."""
        for key in list(recipients.keys()):
            r = recipients[key]
            if not (
                r.data.get("preferences", {})
                .get("notifications", {})
                .get("enabled", False)
            ):
                del recipients[key]

        return recipients



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User roles/groups resource config."""


import marshmallow as ma
from invenio_records_resources.resources import (
    RecordResourceConfig,
    SearchRequestArgsSchema,
)


#
# Request args
#
class GroupSearchRequestArgsSchema(SearchRequestArgsSchema):
    """Add parameter to parse tags."""

    title = ma.fields.String()


#
# Resource config
#
class GroupsResourceConfig(RecordResourceConfig):
    """User groups resource configuration."""

    blueprint_name = "groups"
    url_prefix = "/groups"
    routes = {
        "list": "",
        "item": "/<id>",
        "item-avatar": "/<id>/avatar.svg",
    }

    request_view_args = {
        "id": ma.fields.Str(),
    }

    request_search_args = GroupSearchRequestArgsSchema



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users service configuration."""

from invenio_records_resources.services import RecordServiceConfig, pagination_links
from invenio_records_resources.services.base.config import (
    ConfiguratorMixin,
    FromConfig,
    FromConfigSearchOptions,
    SearchOptionsMixin,
)
from invenio_records_resources.services.records.config import SearchOptions
from invenio_records_resources.services.records.params import (
    FacetsParam,
    QueryStrParam,
    SortParam,
)
from invenio_records_resources.services.records.queryparser import (
    QueryParser,
    SearchFieldTransformer,
)

from ...records.api import UserAggregate
from ..common import Link
from ..params import FixedPagination
from ..permissions import UsersPermissionPolicy
from ..schemas import UserSchema
from ..users import facets
from .results import UserItem, UserList
from .search_params import ModerationFilterParam


class UserSearchOptions(SearchOptions, SearchOptionsMixin):
    """Search options."""

    pagination_options = {
        "default_results_per_page": 10,
        "default_max_results": 10,
    }
    query_parser_cls = QueryParser.factory(
        tree_transformer_cls=SearchFieldTransformer,
        fields=["username^2", "email^2", "profile.full_name^3", "profile.affiliations"],
        allow_list=["username", "email"],  # mapped fields are added on the query parser
        mapping={
            "affiliation": "profile.affiliations",
            "affiliations": "profile.affiliations",
            "full_name": "profile.full_name",
            "fullname": "profile.full_name",
            "name": "profile.full_name",
        },
    )

    params_interpreters_cls = [
        QueryStrParam,
        SortParam,
        FixedPagination,
        FacetsParam,
        ModerationFilterParam.factory(param="is_blocked", field="blocked_at"),
        ModerationFilterParam.factory(param="is_verified", field="verified_at"),
        ModerationFilterParam.factory(param="is_active", field="active"),
    ]

    facets = {
        "email_domain": facets.email_domain,
        "affiliations": facets.affiliations,
    }


class UsersServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Requests service configuration."""

    # common configuration
    permission_policy_cls = UsersPermissionPolicy
    result_item_cls = UserItem
    result_list_cls = UserList
    search = FromConfigSearchOptions(
        "USERS_RESOURCES_SEARCH",
        "USERS_RESOURCES_SORT_OPTIONS",
        "USERS_RESOURCES_SEARCH_FACETS",
        search_option_cls=UserSearchOptions,
    )
    # search = UserSearchOptions
    # specific configuration
    service_id = "users"
    record_cls = UserAggregate
    schema = FromConfig("USERS_RESOURCES_SERVICE_SCHEMA", UserSchema)
    indexer_queue_name = "users"
    index_dumper = None

    # links configuration
    links_item = {
        "self": Link("{+api}/users/{id}"),
        "avatar": Link("{+api}/users/{id}/avatar.svg"),
        # TODO missing moderation actions based on permissions
    }
    links_search = pagination_links("{+api}/users{?args*}")

    components = [
        # order of components are important!
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User specific generators for notifications."""


from invenio_notifications.models import Recipient
from invenio_notifications.services.generators import RecipientGenerator
from invenio_records.dictutils import dict_lookup


class UserRecipient(RecipientGenerator):
    """User recipient generator for a notification."""

    def __init__(self, key):
        """Ctor."""
        self.key = key

    def __call__(self, notification, recipients):
        """Update required recipient information and add backend id."""
        user = dict_lookup(notification.context, self.key)
        recipients[user["id"]] = Recipient(data=user)
        return recipients



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Users and user groups permissions."""

from invenio_records_permissions import BasePermissionPolicy
from invenio_records_permissions.generators import (
    AdminAction,
    AnyUser,
    AuthenticatedUser,
    SystemProcess,
)

from invenio_users_resources.permissions import user_management_action

from .generators import IfGroupNotManaged, IfPublicEmail, IfPublicUser, Self

UserManager = AdminAction(user_management_action)


class UsersPermissionPolicy(BasePermissionPolicy):
    """Permission policy for users and user groups."""

    can_create = [SystemProcess()]
    can_read = [
        UserManager,
        IfPublicUser(then_=[AnyUser()], else_=[Self()]),
        SystemProcess(),
    ]
    can_search = [AuthenticatedUser(), SystemProcess()]
    can_update = [SystemProcess()]
    can_delete = [SystemProcess()]

    can_read_email = [
        UserManager,
        IfPublicEmail([AnyUser()], [Self()]),
        SystemProcess(),
    ]
    can_read_details = [UserManager, Self(), SystemProcess()]

    # Moderation permissions
    can_manage = [UserManager, SystemProcess()]
    can_search_all = [UserManager, SystemProcess()]
    can_read_system_details = [UserManager, SystemProcess()]


class GroupsPermissionPolicy(BasePermissionPolicy):
    """Permission policy for users and user groups."""

    can_create = [SystemProcess()]
    can_read = [
        IfGroupNotManaged([AuthenticatedUser()], [UserManager]),
        SystemProcess(),
    ]
    can_search = [AuthenticatedUser(), SystemProcess()]
    can_update = [SystemProcess()]
    can_delete = [SystemProcess()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User service tests."""

import time

import pytest
from invenio_access.permissions import system_identity
from invenio_cache.lock import CachedMutex
from invenio_cache.proxies import current_cache
from invenio_records_resources.services.errors import PermissionDeniedError

from invenio_users_resources.proxies import current_actions_registry


@pytest.fixture(scope="function", autouse=True)
def mock_action_registry(monkeypatch, user_service):
    """Mocks action to registry entirely.

    Monkeypatches the registry to be modified only for testing.
    """

    for key in current_actions_registry:
        monkeypatch.setitem(current_actions_registry, key, [])
    return True


#
# Search
#
def test_search_restricted(user_service, anon_identity, user_pub):
    """Only authenticated users can search."""
    # Anon identity
    pytest.raises(
        # TODO: Should be mapped to a 404
        PermissionDeniedError,
        user_service.search,
        anon_identity,
    )
    # Authenticated identity
    res = user_service.search(user_pub.identity).to_dict()
    assert res["hits"]["total"] > 0


def test_search_public_users(user_service, user_pub):
    """Only public users are shown in search."""
    res = user_service.search(user_pub.identity).to_dict()
    assert res["hits"]["total"] == 2  # 2 public users in conftest


@pytest.mark.parametrize(
    "query",
    [
        "email:res@inveniosoftware.org",
        "res@inveniosoftware.org",
        "email:pubres@inveniosoftware.org",
        "pubres@inveniosoftware.org",
        "Plazi",
        "+name:Jose -affiliation:CERN",
        "name:Jose AND NOT affiliation:CERN",
        "username:inactive",
        "username:unconfirmed",
        "preferences.visibility:public",
        "preferences.email_visibility:restricted",
        "profile.affiliations:Plazi",
        "invalid:test",
    ],
)
def test_search_field_not_searchable(user_service, user_pub, query):
    """Make sure certain fields are NOT searchable."""
    res = user_service.search(user_pub.identity, q=query).to_dict()
    assert res["hits"]["total"] == 0


@pytest.mark.parametrize(
    "query",
    [
        "affiliations:CERN",
        "affiliation:CERN",
        "name:Jose affiliation:CERN",
        "+name:Jose +affiliation:CERN",
        "CERN",
        "Tim",
        "Tim CERN",
        "Jose",
        "Jose CERN",
        "email:pub@inveniosoftware.org",
        "username:pub",
    ],
)
def test_search_field(user_service, user_pub, query):
    """Make sure certain fields ARE searchable."""
    res = user_service.search(user_pub.identity, q=query).to_dict()
    assert res["hits"]["total"] > 0


#
# Read
#
def test_read_with_anon(user_service, anon_identity, user_pub, user_pubres, user_res):
    """Anonymous users can read a single *public* user."""
    res = user_service.read(anon_identity, user_pub.id).to_dict()
    assert res["username"] == "pub"
    assert res["email"] == user_pub.email
    res = user_service.read(anon_identity, user_pubres.id).to_dict()
    assert res["username"] == "pubres"
    assert "email" not in res
    pytest.raises(
        # TODO: Should be mapped to a 404
        PermissionDeniedError,
        user_service.read,
        anon_identity,
        user_res.id,
    )


@pytest.mark.parametrize(
    "username,can_read", [("pub", True), ("pubres", True), ("res", False)]
)
def test_read_avatar_with_anon(user_service, anon_identity, users, username, can_read):
    """Anonymous users can read avatar single *public* user."""
    user = users[username]
    if can_read:
        user_service.read_avatar(anon_identity, user.id)
    else:
        pytest.raises(
            # TODO: Should be mapped to a 404
            PermissionDeniedError,
            user_service.read_avatar,
            anon_identity,
            user.id,
        )


@pytest.mark.parametrize(
    "username",
    [
        "pub",
        "pubres",
        "res",
    ],
)
def test_read_self(user_service, users, username):
    """Users can read themselves."""
    user = users[username]
    res = user_service.read(user.identity, user.id).to_dict()
    assert res["username"] == username
    # Email can be seen even if restricted
    assert res["email"] == user.email
    # Avatar can also be seen.
    user_service.read_avatar(user.identity, user.id)


def test_search_permissions(app, db, user_service, user_moderator, user_res):
    """Test service search for permissions."""
    # User can search for himself
    search = user_service.search(
        user_res.identity, q=f"username:{user_res._user.username}"
    )
    assert search.total > 0

    # User can't search for non-confirmed users
    with pytest.raises(PermissionDeniedError):
        user_service.search_all(user_res.identity, user_res.id)

    # Moderator can search for any user
    search = user_service.search_all(
        user_moderator.identity, q=f"username:{user_res._user.username}"
    )
    assert search.total > 0


def test_block(
    app, db, user_service, user_moderator, user_res, clear_cache, search_clear
):
    """Test user block."""

    with pytest.raises(PermissionDeniedError):
        user_service.block(user_res.identity, user_res.id)

    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked

    ur = user_service.read(user_res.identity, user_res.id)
    # User can't see when it was blocked
    assert not "blocked_at" in ur.data
    # But can see it's not active
    assert ur.data["active"] is False

    ur = user_service.read(user_moderator.identity, user_res.id)
    # Moderator can see the blocked_at time
    assert ur.data["blocked_at"] is not None

    # TODO user is blocked, the user should not be able to search.
    # search = user_service.search(user_res.identity, q=f"username:{ur._user.username}")
    # assert search.total == 0

    # Moderator can still search for the user
    search = user_service.search_all(
        user_moderator.identity, q=f"username:{user_res._user.username}"
    )
    assert search.total > 0


def test_approve(
    app, db, user_service, user_res, user_moderator, clear_cache, search_clear
):
    """Test approval of an user."""
    with pytest.raises(PermissionDeniedError):
        user_service.block(user_res.identity, user_res.id)

    approved = user_service.approve(user_moderator.identity, user_res.id)
    assert approved

    ur = user_service.read(user_res.identity, user_res.id)
    # User can't see when it was approved
    assert not "verified_at" in ur.data
    # But can see it's active
    assert ur.data["active"] is True

    ur = user_service.read(user_moderator.identity, user_res.id)
    # Moderator can see when it was approved
    assert "verified_at" in ur.data


def test_deactivate(app, db, user_service, user_res, user_moderator, clear_cache):
    """Test deactivation of an user."""
    with pytest.raises(PermissionDeniedError):
        user_service.block(user_res.identity, user_res.id)

    deactivated = user_service.deactivate(user_moderator.identity, user_res.id)
    assert deactivated

    ur = user_service.read(user_res.identity, user_res.id)
    # User can see it's not active
    assert ur.data["active"] is False

    # Moderator can still search for the user
    search = user_service.search_all(
        user_moderator.identity, q=f"username:{user_res._user.username}"
    )
    assert search.total > 0


def test_non_existent_user_management(app, db, user_service, user_moderator):
    """Try to manage a non-existent user."""
    fake_user_id = 1000
    funcs = [
        user_service.block,
        user_service.approve,
        user_service.deactivate,
        user_service.restore,
    ]
    for f in funcs:
        with pytest.raises(PermissionDeniedError):
            f(user_moderator.identity, fake_user_id)


def test_restore(app, db, user_service, user_res, user_moderator, clear_cache):
    """Test restore of a user."""
    blocked = user_service.block(user_moderator.identity, user_res.id)
    assert blocked

    ur = user_service.read(user_moderator.identity, user_res.id)
    assert ur.data["active"] == False
    assert ur.data["blocked_at"] is not None

    restored = user_service.restore(user_moderator.identity, user_res.id)
    assert restored

    ur = user_service.read(user_moderator.identity, user_res.id)
    assert ur.data["active"] == True
    assert ur.data["verified_at"] is not None
    assert not "blocked_at" in ur.data


# TODO Clear the cache to test actions without locking side-effects



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module providing management APIs for users and roles/groups."""

from invenio_i18n import lazy_gettext as _

from invenio_users_resources.services.schemas import UserSchema
from invenio_users_resources.services.users import facets

USERS_RESOURCES_AVATAR_COLORS = [
    "#e06055",
    "#ff8a65",
    "#e91e63",
    "#f06292",
    "#673ab7",
    "#ba68c8",
    "#7986cb",
    "#3f51b5",
    "#5e97f6",
    "#00a4e4",
    "#4dd0e1",
    "#0097a7",
    "#d4e157",
    "#aed581",
    "#57bb8a",
    "#4db6ac",
    "#607d8b",
    "#795548",
    "#a1887f",
    "#fdd835",
    "#a3a3a3",
    "#556c60",
    "#605264",
    "#923035",
    "#915a30",
    "#55526f",
    "#67635a",
]

USERS_RESOURCES_SERVICE_SCHEMA = UserSchema
"""Schema used by the users service."""

USERS_RESOURCES_SEARCH = {
    "sort": [
        "email",
        "username",
        "email_domain",
    ],
    "facets": ["email_domain", "affiliations"],
}
"""User search configuration (i.e list of banners)."""

USERS_RESOURCES_SORT_OPTIONS = {
    "username": dict(
        title=_("Username"),
        fields=["username"],
    ),
    "email": dict(
        title=_("Email"),
        fields=["email"],
    ),
    "email_domain": dict(
        title=_("Email domain"),
        fields=["email.domain"],
    ),
}
"""Definitions of available Users sort options. """

USERS_RESOURCES_SEARCH_FACETS = {
    "email_domain": {
        "facet": facets.email_domain,
        "ui": {
            "field": "email.domain",
        },
    },
    "affiliations": {
        "facet": facets.affiliations,
        "ui": {
            "field": "profile.affiliations.keyword",
        },
    },
}
"""Invenio requests facets."""

USERS_RESOURCES_ADMINISTRATION_ENABLED = False
"""Enable the user administration"""

USERS_RESOURCES_MODERATION_LOCK_DEFAULT_TIMEOUT = 30
"""Default timeout, in seconds, to lock a user when moderating."""

USERS_RESOURCES_MODERATION_LOCK_RENEWAL_TIMEOUT = 120
"""Renewal timeout, in seconds, to increase the lock time for a user when moderating."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-users-resources is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Implements locking for users."""

from flask import current_app
from invenio_cache.lock import CachedMutex
from werkzeug.local import LocalProxy

timeout_default = LocalProxy(
    lambda: current_app.config.get("USERS_RESOURCES_MODERATION_LOCK_DEFAULT_TIMEOUT")
)


class ModerationMutex(CachedMutex):
    """Wrapper of ``CachedMutex`` to be used solely in user moderation.

    This class forces the lock ID prefix and specifies a default timeout when acquiring the lock.
    """

    lock_id_prefix = "user_moderation_lock"

    def __init__(self, user_id):
        """Constructor.

        Creates a lock using the parent constructor, building the lock ID as follows:

            <id_prefix>.<user_id>
        """
        super().__init__(lock_id=f"{self.lock_id_prefix}.{user_id}")

    def acquire(self, timeout=None):
        """Acquires the lock.

        If the timeout is not provided, a default is retrieved from the config USERS_RESOURCES_MODERATION_LOCK_DEFAULT_TIMEOUT
        """
        timeout = timeout or timeout_default
        super().acquire(timeout=timeout_default)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Base model classes for user and group management in Invenio."""

from abc import ABC, abstractmethod

from invenio_accounts.proxies import current_datastore


class MockModel(dict, ABC):
    """Model class that does not correspondond to a database table.

    Since we already have all information about the required entities
    stored in the database and just want to provide a central API to
    manage these entities as one, we do not want to store any of the
    information in the database again.
    Instead, we provide a mock model that provides the model interface
    that's expected by API classes as far as necessary.
    """

    def __init__(self, data=None, model_obj=None, **kwargs):
        """Constructor."""
        data.update(kwargs)
        super().__init__(data)
        self._model_obj = model_obj

    @property
    @abstractmethod
    def model_obj(self):
        """The actual model object behind this mock model."""
        return None

    @property
    def id(self):
        """User identifier."""
        return self.data["id"]

    @property
    def created(self):
        """When the user was created."""
        return self.model_obj.created

    @property
    def updated(self):
        """When the user was last updated."""
        return self.model_obj.updated

    @property
    def version_id(self):
        """Used by SQLAlchemy for optimistic concurrency control."""
        return self.model_obj.version_id

    @property
    def data(self):
        """Get the user's data by decoding the JSON."""
        return dict(self)

    @property
    def json(self):
        """Provide the user's data as a JSON/dict blob."""
        return dict(self)

    @property
    def is_deleted(self):
        """Boolean flag to determine if a user is soft deleted."""
        # TODO
        return False

    @property
    def blocked_at(self):
        """Date when the user was blocked, if any."""
        return self.model_obj.blocked_at

    @blocked_at.setter
    def blocked_at(self, value):
        self.model_obj.blocked_at = value

    @property
    def verified_at(self):
        """Date when the user was verified, if any."""
        return self.model_obj.verified_at

    @verified_at.setter
    def verified_at(self, value):
        self.model_obj.verified_at = value

    @property
    def active(self):
        """Boolean flag to determine if a user is active."""
        return self.model_obj.active

    @active.setter
    def active(self, value):
        self.model_obj.active = value


class UserAggregateModel(MockModel):
    """Mock model for glueing together various parts of user info."""

    @property
    def model_obj(self):
        """The actual model object behind this mock model."""
        if self._model_obj is None:
            id_ = self.data.get("id")
            email = self.data.get("email")
            self._model_obj = current_datastore.get_user(id_ or email)
        return self._model_obj


class GroupAggregateModel(MockModel):
    """Mock model for glueing together various parts of user group info."""

    @property
    def model_obj(self):
        """The actual model object behind this mock model."""
        if self._model_obj is None:
            name = self.data.get("id")
            self._model_obj = current_datastore.find_role(name)
        return self._model_obj



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Avatar results for users and groups."""

from datetime import datetime, timedelta
from io import BytesIO

from flask import render_template


class AvatarResult:
    """An avatar result for a user or group."""

    def __init__(
        self,
        user_or_group,
    ):
        """Constructor."""
        self._obj = user_or_group

    @property
    def bytes_io(self):
        """Get the avatar image file."""
        return BytesIO(
            render_template(
                "avatar.svg",
                bg_color=self._obj.avatar_color,
                text=self._obj.avatar_chars,
            ).encode("utf8")
        )

    @property
    def mimetype(self):
        """Get the MIME type."""
        return "image/svg+xml"

    @property
    def name(self):
        """Get the filename of the avatar."""
        return "avatar.svg"

    @property
    def etag(self):
        """Get an ETag for the avatar."""
        return f"{self._obj.avatar_chars}{self._obj.avatar_color}"

    @property
    def last_modified(self):
        """Get last modified date for the response."""
        max_age = datetime.utcnow() - timedelta(days=7)
        return self._obj.updated if self._obj.updated > max_age else max_age

    @property
    def max_age(self):
        """Get time out duration for cached avatars in seconds."""
        # As avatars are often called it should be cached to reduce the load on the user
        # currently set to 5 minutes as a reasonable time
        return 60 * 5
